@@ -7,13 +7,26 @@
 
 #include <NTL/ALL_FEATURES.h>
 
+#include <NTL/PackageInfo.h>
 
-// defines the working C++ standard
+#if (defined(__GNUC__) && (defined(__i386__) || defined(__x86_64__)))
+#define NTL_GNUC_INTEL
+#endif
 
-#if defined(NTL_STD_CXX11)
-#define NTL_CXX_STANDARD (2011)
-#elif defined(NTL_STD_CXX14)
+#if (!defined(NTL_HAVE_LL_TYPE) && defined(NTL_WINPACK) &&  (defined(_MSC_VER) || defined(NTL_GNUC_INTEL)))
+// for the windows distribution, 
+//   we assume LL_TYPE works for MSVC++ (which is true for both x86 and ARM)
+//   and for GNUC/Intel platforms (e.g., Code Blocks)
+#define NTL_HAVE_LL_TYPE
+#endif
+
+// Define the working C++ standard.
+// Both NTL_STD_CXX14 and NTL_STD_CXX11, and we take the highest one
+
+#if defined(NTL_STD_CXX14)
 #define NTL_CXX_STANDARD (2014)
+#elif defined(NTL_STD_CXX11)
+#define NTL_CXX_STANDARD (2011)
 #else
 #define NTL_CXX_STANDARD (1998)
 #endif
@@ -293,7 +306,7 @@ extern unsigned long exception_counter;
 
 #define NTL_ULONG_TO_LONG(a) \
    ((((unsigned long) a) >> (NTL_BITS_PER_LONG-1)) ? \
-    (((long) (((unsigned long) a) - ((unsigned long) NTL_MIN_LONG))) + \
+    (((long) (((unsigned long) a) ^ ((unsigned long) NTL_MIN_LONG))) ^ \
        NTL_MIN_LONG) : \
     ((long) a))
 
@@ -308,7 +321,7 @@ extern unsigned long exception_counter;
 
 #define NTL_UINT_TO_INT(a) \
    ((((unsigned int) a) >> (NTL_BITS_PER_INT-1)) ? \
-    (((int) (((unsigned int) a) - ((unsigned int) NTL_MIN_INT))) + \
+    (((int) (((unsigned int) a) ^ ((unsigned int) NTL_MIN_INT))) ^ \
        NTL_MIN_INT) : \
     ((int) a))
 
@@ -347,6 +360,9 @@ extern unsigned long exception_counter;
 
 
 
+double _ntl_GetWallTime();
+
+
 long _ntl_IsFinite(double *p);
 /* This forces a double into memory, and tests if it is "normal";
    that means, not NaN, not +/- infinity, not denormalized, etc.
@@ -493,6 +509,11 @@ char *_ntl_make_aligned(char *p, long align)
 
 #define NTL_AVX_LOCAL_ARRAY(x, type, n) NTL_ALIGNED_LOCAL_ARRAY(NTL_AVX_BYTE_ALIGN, x, type, n)
 
+#define NTL_AVX512_BYTE_ALIGN (64)
+
+#define NTL_AVX512_LOCAL_ARRAY(x, type, n) NTL_ALIGNED_LOCAL_ARRAY(NTL_AVX512_BYTE_ALIGN, x, type, n)
+
+
 #define NTL_DEFAULT_ALIGN (64)
 // this should be big enough to satisfy any SIMD instructions,
 // and it should also be as big as a cache line
@@ -562,6 +583,10 @@ _ntl_bpl_divrem(unsigned long a, long& q, long& r)
 }
 
 
+// vectors are grown by a factor of 1.5
+inline long _ntl_vec_grow(long n)
+{ return n + n/2; }
+
 
 template <class T>
 struct _ntl_is_char_pointer
@@ -594,5 +619,4 @@ struct _ntl_enable_if<true, T> {
 
 
 
-
 #endif
