@@ -481,22 +481,15 @@ void mul(ZZ_pX& c, const ZZ_pX& a, const ZZ_pX& b)
       long mbits;
       mbits = NumBits(ZZ_p::modulus());
 
-      long nt = AvailableThreads();
-      // right now, SSMul does not parallelize very well
+      // long nt = AvailableThreads();
+      // SSMul is now fully thread boosted
       
       double rat = SSRatio(deg(a), mbits, deg(b), mbits);
 
-      if ( nt == 1 && (
+      if ( (k >= 106 && rat < 1.50) || 
+           (k >= 212 && rat < 1.75) ) {
 
-         (k >= 106 && rat < 1.50) || 
-         (k >= 212 && rat < 1.75) 
-
-      )) {
-         ZZX A, B, C;
-         conv(A, a);
-         conv(B, b);
-         SSMul(C, A, B);
-         conv(c, C);
+         SSMul(c, a, b);
       }
       else {
          FFTMul(c, a, b);
@@ -532,28 +525,15 @@ void sqr(ZZ_pX& c, const ZZ_pX& a)
       mbits = NumBits(ZZ_p::modulus());
 
 
-      long nt = 1;
-      // FIXME: needs to be updated when I thread-enable the SS
-      // mul routine
-
-#ifdef NTL_THREAD_BOOST
-      BasicThreadPool *pool = GetThreadPool();
-      if (pool && !pool->active()) nt  = pool->NumThreads();
-#endif
+      // long nt = AvailableThreads();
+      // SSMul is now fully thread boosted
 
       double rat = SSRatio(deg(a), mbits, deg(a), mbits);
 
-      if ( nt == 1 && (
+      if ( (k >= 53  && rat < 1.20) || (k >= 106 && rat < 1.30) || 
+           (k >= 212 && rat < 1.75) ) {
 
-         (k >= 53  && rat < 1.20) || 
-         (k >= 106 && rat < 1.30) || 
-         (k >= 212 && rat < 1.75) 
-
-      )) {
-         ZZX A, C;
-         conv(A, a);
-         SSSqr(C, A);
-         conv(c, C);
+         SSSqr(c, a);
       }
       else {
          FFTSqr(c, a);
@@ -1862,7 +1842,7 @@ NTL_TBDECL(FromFFTRep)(ZZ_pX& x, FFTRep& y, long lo, long hi)
    l = max(l, 0);
 
    long len = y.len;
-   if (len <= hi) LogicError("FromFFTRep: bad len");
+   if (len <= hi) LogicError("FromFFTRep: bad len 1"); 
 
 
    for (i = 0; i < nprimes; i++) {
@@ -1912,7 +1892,7 @@ void FromFFTRep(ZZ_pX& x, FFTRep& y, long lo, long hi)
    l = max(l, 0);
 
    long len = y.len;
-   if (len <= hi) LogicError("FromFFTRep: bad len");
+   if (len <= hi) LogicError("FromFFTRep: bad len 2");
 
    pool->exec_range(nprimes,
    [&y, k, len](long first, long last) {
@@ -2095,7 +2075,7 @@ NTL_TBDECL(NDFromFFTRep)(ZZ_pX& x, const FFTRep& y, long lo, long hi, FFTRep& z)
    l = max(l, 0);
 
    long len = y.len;
-   if (len <= hi) LogicError("FromFFTRep: bad len");
+   if (len <= hi) LogicError("FromFFTRep: bad len 3");
 
    z.SetSize(k);
 
@@ -2147,7 +2127,7 @@ void NDFromFFTRep(ZZ_pX& x, const FFTRep& y, long lo, long hi, FFTRep& z)
    l = max(l, 0);
 
    long len = y.len;
-   if (len <= hi) LogicError("FromFFTRep: bad len");
+   if (len <= hi) LogicError("FromFFTRep: bad len 4");
 
    z.SetSize(k);
 
@@ -2219,7 +2199,7 @@ NTL_TBDECL(FromFFTRep)(ZZ_p* x, FFTRep& y, long lo, long hi)
    n = (1L << k);
 
    //if (y.len <= min(hi, n-1)) LogicError("FromFFTRep: bad len");
-   if (y.len != n) LogicError("FromFFTRep: bad len");
+   if (y.len != n) LogicError("FromFFTRep: bad len 5");
 
    long nprimes = FFTInfo->NumPrimes;
    t.SetLength(nprimes);
@@ -2267,7 +2247,7 @@ void FromFFTRep(ZZ_p* x, FFTRep& y, long lo, long hi)
    n = (1L << k);
 
    //if (y.len <= min(hi, n-1)) LogicError("FromFFTRep: bad len");
-   if (y.len != n) LogicError("FromFFTRep: bad len");
+   if (y.len != n) LogicError("FromFFTRep: bad len 6");
 
    long nprimes = FFTInfo->NumPrimes;
 
@@ -2889,7 +2869,7 @@ void FromFFTRep(ZZ_pXModRep& x, const FFTRep& a)
    long k = a.k;
    long n = 1L << k;
 
-   if (a.len != n) LogicError("FromFFTRep: bad len");
+   if (a.len != n) LogicError("FromFFTRep: bad len 7");
 
    x.SetSize(n);
    for (long i = 0; i < nprimes; i++) {
